/***************************************************************************
 *   Copyright (C) 2012 by Laboratoire d'Economie Forestière               *
 *   http://www.nancy.inra.fr/lef_eng                                      *
 *   Author: Antonello Lobianco - antonello@lobianco.org                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include <iostream>
#include <fstream>
#include <sstream>
#include <cstdlib> // for exit function
#include <math.h>

#include <vector>
#include <string>
#include <map>
#include <algorithm>

#include "anyoption.h"


using namespace std;


struct databydim1{
	string dim1;
	multimap<double,string> probmap;
};

struct result {
	string dim1;
	string dim2;
};

struct idata {
	string dim1;
	string dim2;
	double prob;
};

// overloading of the "()" operator to define a compare function for my idata structure needed by the sort algorithm
struct idata_less {
	bool operator ()(idata const& a, idata const& b) const {
		// first ordering by first dimension
		if (a.dim1 < b.dim1) return true;
		if (a.dim1 > b.dim1) return false;
		// only if the first dimension is the same, sorting by prob
		if (a.prob < b.prob) return true;
		return false;
	}
};


void loadData(string filename);
void sample(int snum);
void printOutput(string filename);
vector <databydim1> aggregatedData;
vector <result> results;
vector <idata> inputdata;
string dim1_header;
string dim2_header;
char fieldDelimiter;
bool inputFile_has_headers;

/**
Version 3
Generalised case

Version 2
This version sort first the data in order of probability for each individual.
All visitor<->forests combinations have now the possibility of being sampled.
*/


int main(int argc, char *argv[]){
	cout << endl;
	cout << "*******************************************************************" << endl;
	cout << "***                     customSampling 2D                       ***" << endl;
	cout << "***         Compiled on: " << __DATE__ << " - " << __TIME__ << "                 ***" << endl;
	cout << "*******************************************************************" << endl<<endl;

	cout <<"Sampling records from two dimensions with a custom probability distribution over the second dimension." << endl;

	int sampleNumber;
	string ifilename;
	string ofilename;
	dim1_header = "Dim1";
	dim2_header = "Dim2";


		// 1. CREATE AN OBJECT
	AnyOption *opt = new AnyOption();


	// 2. SET PREFERENCES
	//opt->setVerbose(); // print warnings about unknown options
	//opt->autoUsagePrint(true); // print usage for bad options

	// 3. SET THE USAGE/HELP
	opt->addUsage( "*** customSampling 2D ***" );
	opt->addUsage( "Usage: " );
	opt->addUsage( "" );
	opt->addUsage( " -h  --help			Prints this help " );
	opt->addUsage( " -i  --input_file [input_file_name]	Input file " );
	opt->addUsage( " -o  --output_file [output_file_name]	Output file (relative path, default: 'sampling_results.txt') " );
	opt->addUsage( " -d  --field_delimiter [delimiter]	Field delimiter (default: tab) " );
	opt->addUsage( " -s  --sample [number_of_sample]	Number of items to sample for each element of the first dimension ");
	opt->addUsage( " -n  --noheaders			Load data starting directly from the first row (default: false) " );
	opt->addUsage( "" );
	opt->addUsage( "Notes:");
	opt->addUsage( "	- sorting of probabilities and usage of the Kahan summation algorithm for very precise results;" );
	opt->addUsage( "	- handle small probabilities up to 1.7E-308;" );
	opt->addUsage( "	- very fast: 15 seconds on over 2M records;" );
	opt->addUsage( "" );
	opt->addUsage( "Limitations:");
	opt->addUsage( "	- input file must be UTF;" );
	opt->addUsage( "	- input file must use English locale (dots as decimal separator);" );
	opt->addUsage( "	- fields in the input file must not be surrounded by quotes;" );
	opt->addUsage( "	- output file uses a single \\n as line terminator. This shoudn't be a problem';" );
	opt->addUsage( "" );
	opt->addUsage( "Report any bug to antonello@lobianco.org." );
	opt->addUsage( "" );

	// 4. SET THE OPTION STRINGS/CHARACTERS
	opt->setFlag(  "help", 'h' );
	opt->setOption(  "input_file", 'i' );
	opt->setOption(  "output_file", 'o' );
	opt->setOption(  "field_delimiter", 'd' );
	opt->setOption(  "sample", 's' );
	opt->setOption(  "noheaders", 'n' );

	// 5. PROCESS THE COMMANDLINE
	opt->processCommandArgs( argc, argv );

	// 6. GET THE VALUES
	if( opt->getFlag( "help" ) || opt->getFlag( 'h' ) || opt->getArgc() >0 ) {
		opt->printUsage();
		delete opt;
		return EXIT_FAILURE;
	}

	if( opt->getValue( 'i' ) != NULL  || opt->getValue( "input_file" ) != NULL  ){
		ifilename = opt->getValue( 'i' );
	}
	else {
		cerr << "***ERROR: Missing parameter: input file" << endl;
		opt->printUsage();
		delete opt;
		return EXIT_FAILURE;
	}

	if( opt->getValue( 'o' ) != NULL  || opt->getValue( "output_file" ) != NULL  ){
		ofilename = opt->getValue( 'o' );
	}
	else {
		ofilename = "sampling_results.txt";
	}

	if( opt->getValue( 'd' ) != NULL  || opt->getValue( "field_delimiter" ) != NULL  ){
		fieldDelimiter = *(opt->getValue( 'd' ));
	}
	else {
		fieldDelimiter = '\t';
	}

	if( opt->getValue( 's' ) != NULL  || opt->getValue( "sample" ) != NULL  ){
		//string sample = opt->getValue( 's' );
		stringstream ss(opt->getValue( 's' ));
		ss >> sampleNumber;
	}
	else {
		cerr << "***ERROR: Missing parameter: sample" << endl;
		opt->printUsage();
		delete opt;
		return EXIT_FAILURE;
	}

	if( opt->getFlag( 'n' ) || opt->getFlag( "noheaders" ) ){
		inputFile_has_headers = false;
	} else {
		inputFile_has_headers = true;
	}

	loadData(ifilename);
	sample(sampleNumber);
	printOutput(ofilename);

	cout << endl<<"**** Sampling report ****" << endl;
	cout << "All done." << endl;
	cout << "Input file used:\t" << ifilename << endl;
	cout << "Sampled pairs saved to:\t" << ofilename << endl;
	cout << "Found " << aggregatedData.size() << " elements of "<< dim1_header << " for which we sampled " << sampleNumber << " " << dim2_header << " each."<<endl<<endl;

	return 0;
}

void loadData(string filename) {
	cout << "Loading input.." << endl;
	ifstream inFile (filename.c_str(), ios::in);
	if(!inFile) { // file couldn't be opened
		cerr << "Error: input file could not be opened" << endl;
		exit(1);
	}
	string line;
	int linecount = 0;
	string dim1id, dim2id;
	double prob;

	while (getline (inFile, line)) {
		istringstream linestream(line);
		if(!linestream) continue; // empty line, go to next loop
		string item;
		int itemnum = 0;

		if(linecount == 0 && inputFile_has_headers) {
			linecount ++;
			// saving headers name..
			while (getline (linestream, item, fieldDelimiter)){
				itemnum++;
				if(itemnum==1) {
					dim1_header = item ;
				} else if (itemnum ==2) {
					dim2_header = item ;
				}
			}
			continue; // skip and go to the next loop if we are on the first line and we have an input file with headers
		}


		while (getline (linestream, item, fieldDelimiter)){
			itemnum++;
			if(itemnum==1) {
				dim1id = item ;
			} else if (itemnum ==2) {
				dim2id = item ;
			} else {
				stringstream ss(item);
				ss >> prob;
			}
		}
		linecount ++;
		idata thislinedata;
		thislinedata.dim1 = dim1id;
		thislinedata.dim2 = dim2id;
		thislinedata.prob = prob;
		inputdata.push_back(thislinedata);
	}
	inFile.close();

	cout << "Sorting input data by " << dim1_header << " and then prob.." << endl;
	sort(inputdata.begin(), inputdata.end(), idata_less());

	cout << "Aggregating data.." << endl;
	string last_dim1id="oiejrtlkj43k5j3490dkfdlkkèpp324234TTTTTY,ò.23671-.3wer"; // difficult you have this id in yr data ;-)
	double cumProb=0.;
	double correction = 0.;
	databydim1* dim1Element;
	for (unsigned int i=0;i<inputdata.size();i++){
		dim1id = inputdata[i].dim1;
		dim2id = inputdata[i].dim2;
		prob = inputdata[i].prob;
		if (dim1id == last_dim1id){
			// using Kahan summation algorithm
			// http://en.wikipedia.org/wiki/Kahan_summation_algorithm
			// http://ideone.com/iuPPP
			double y = prob-correction;
			double t = cumProb + y;
			correction = (t-cumProb)-y;
			cumProb = t;
			//cumProb += prob; dump summation
			dim1Element->probmap.insert(pair<double, string >(cumProb, dim2id));
		} else {
			if(last_dim1id != "oiejrtlkj43k5j3490dkfdlkkèpp324234TTTTTY,ò.23671-.3wer") aggregatedData.push_back(*dim1Element); // don't do this on the first line
			dim1Element = new databydim1();
			dim1Element->dim1 = dim1id;
			last_dim1id = dim1id;
			cumProb = prob;
			dim1Element->probmap.insert(pair<double, string >(cumProb, dim2id));
		}

	}
	aggregatedData.push_back(*dim1Element);


}

void sample(int snum){
	cout << "Sampling.." << endl;
	result individualresult;

	for(unsigned int i=0;i<aggregatedData.size();i++){
		for(int y=0;y<snum;y++){
			double random = ((double)rand() / ((double)(RAND_MAX)+(double)(1)) ); // [0,1]
			multimap<double,string>::const_iterator it;
			it = aggregatedData[i].probmap.upper_bound(random);
			if (it!=aggregatedData[i].probmap.end()){
				individualresult.dim1 = aggregatedData[i].dim1;
				individualresult.dim2 = it->second;
				results.push_back(individualresult);
			} else {
				cerr << "Opss... found a random number higher than the cumProb for individual "<<aggregatedData[i].dim1<<". Sure the sum of the probability is 1 ??" << endl;
				exit(1);
			}
		}
	}

}

void printOutput(string filename){
	cout << "Printing output.." << endl;
	ofstream outFile (filename.c_str(), ios::out);
	if(!outFile) { // file couldn't be opened
		cerr << "Error: file could not be opened for writing" << endl;
		exit(1);
	}
	outFile << dim1_header << fieldDelimiter << dim2_header << "\n"; // \t is tab character, \n is end of line
	for (unsigned int i=0;i<results.size();i++){
	 outFile << results[i].dim1 << "\t" << results[i].dim2 << "\n";
	}

	outFile.close();

}



